﻿using System;
using System.Collections.Generic;
using System.Text;

namespace SipEmergencySniffer
{
    static class GeoTrans
    {
        /**
        *  Converts from geographic (degrees) to UTM coordinates (meters). The ellipsoid model used for computations is WGS84.
        *  Adapted from code by Charles L. Taylor. Reference: B. Hoffmann-Wellenhof, H. Lichtenegger, and J. Collins,
        *  GPS: Theory and Practice, 3rd ed.  New York: Springer-Verlag Wien, 1994.
        *  
        *  @author Lars Martin Kristensen, Feb 06, 2007. 
        */

        //Ellipsoid model constants (actual values here are for WGS84)
        private static double sm_a = 6378137.0;
        private static double sm_b = 6356752.314;
        private static double sm_EccSquared = 6.69437999013e-03;
        private static double UTMScaleFactor = 0.9996;

        //earths radius, used for great circle distance measurments
        private static double earth_avg_rad = 6372795;

        public static double[] getCoordinatesFromUTM(double x, double y)
        {
            double[] res = UTMXYToLatLon(x, y, 32, false);
            res[0] = (res[0] / Math.PI) * 180;
            res[1] = (res[1] / Math.PI) * 180;
            return res;
        }

        public static double[] getUTMFromCoordinates(double lon, double lat)
        {
            double[] res = latLonToUTMXY(lon, lat);
            return res;
        }

        /**
         *  Find the Great circle distance between two points specified in lon/lat using the Haversine formula
         *  Using a sphere with a radius of 6372795 m which results in an error of up to about 0.5%.
         *  Ref: http://en.wikipedia.org/wiki/Great-circle_distance 
         * 
         * @param lat1	lat of point one in radians
         * @param lon1	lon of point one in radians
         * @param lat2	lat of point two in radians
         * @param lon2	lon of point one in radians
         * @return		the distance in meters
         */
        public static double greatCircleDistance(double lat1, double lon1, double lat2, double lon2)
        {
            double a = Math.Sin((lat2 - lat1) / 2) * Math.Sin((lat2 - lat1) / 2) +
                        Math.Cos(lat1) * Math.Cos(lat2) * Math.Sin((lon2 - lon1) / 2) * Math.Sin((lon2 - lon1) / 2);

            return earth_avg_rad * 2 * Math.Atan2(Math.Sqrt(a), Math.Sqrt(1 - a));
        }


        /**
         * Find the Great circle distance between two points. This is only accurate if the points are very
         * close. However it's fast to compute and and easy to implement om simple platforms (JavaME)
         * 
         * @param lat1	lat of point one in radians
         * @param lon1	lon of point one in radians
         * @param lat2	lat of point two in radians
         * @param lon2	lon of point one in radians
         * @return		the distance in meters
         */
        public static double simpleGreatCircleDistance(double lat1, double lon1, double lat2, double lon2)
        {
            double lonDist = (lon2 - lon1) * earth_avg_rad * Math.Cos((lat1 + lat2) / 2);
            double latDist = (lat2 - lat1) * earth_avg_rad;
            return Math.Sqrt(lonDist * lonDist + latDist * latDist);
        }


        /**
         * Converts from long/lat on nmea-0183 form to degrees
         * 
         * @param nmea  An coordianate fetched from a nmea-0183 string (DDMM.MMM or DDDMM.MMM)
         * @return      The corresponding coordinate in degrees (D*.D*)
         */
        public static double nmeaToDeg(String nmea)
        {
            int cPos = nmea.IndexOf('.');
            return Convert.ToInt16(nmea.Substring(0, cPos - 2)) + Convert.ToDouble(nmea.Substring(cPos - 2)) / 60;
        }


        /**
         * Converts from decimal degrees to degrees, minutes and seconds
         * 
         * @param dec	a decimal degree
         * @return		an int-array of size tree containing dec, min and sec
         */
        public static int[] decToDegMinSec(double dec)
        {
            int[] degMinSec = new int[3];
            degMinSec[0] = (int)dec;
            degMinSec[1] = (int)((dec - degMinSec[0]) * 60);
            degMinSec[2] = (int)((((dec - degMinSec[0]) * 60) - degMinSec[1]) * 60);
            return degMinSec;
        }


        /**
         * Converts from degrees, minutes and seconds to decimal degrees
         * 
         * @param deg	degrees
         * @param min	minutes
         * @param sec	seconds
         * 
         * @return		a decimal degree
         */
        public static double degMinSecToDec(double deg, double min, double sec)
        {
            return deg + (min / 60) + (sec / 3600);
        }

        /**
         * Computes the ellipsoidal distance from the equator to a point at a given latitude.
         * 
         * @param phi   Latitude of the point, in radians.
         * @return      The ellipsoidal distance of the point from the equator, in meters.
         */
        private static double arcLengthOfMeridian(double phi)
        {
            double alpha, beta, gamma, delta, epsilon, n;

            //Precalculations
            n = (sm_a - sm_b) / (sm_a + sm_b);
            alpha = ((sm_a + sm_b) / 2.0) * (1.0 + (Math.Pow(n, 2.0) / 4.0) + (Math.Pow(n, 4.0) / 64.0));
            beta = (-3.0 * n / 2.0) + (9.0 * Math.Pow(n, 3.0) / 16.0) + (-3.0 * Math.Pow(n, 5.0) / 32.0);
            gamma = (15.0 * Math.Pow(n, 2.0) / 16.0) + (-15.0 * Math.Pow(n, 4.0) / 32.0);
            delta = (-35.0 * Math.Pow(n, 3.0) / 48.0) + (105.0 * Math.Pow(n, 5.0) / 256.0);
            epsilon = (315.0 * Math.Pow(n, 4.0) / 512.0);

            //Calculate the sum of the series and return
            return alpha * (phi + (beta * Math.Sin(2.0 * phi)) + (gamma * Math.Sin(4.0 * phi)) + (delta * Math.Sin(6.0 * phi)) + (epsilon * Math.Sin(8.0 * phi)));
        }


        /**
         * Determines the central meridian for the given UTM zone.
         * 
         * @param zone    An integer value designating the UTM zone, range [1,60].
         * @return        The central meridian for the given UTM zone, in radians, or zero if the UTM zone parameter is outside the range [1,60].
         *                Range of the central meridian is the radian equivalent of [-177,+177].
         */
        private static double UTMCentralMeridian(int zone)
        {
            return ToRadians(-183.0 + (zone * 6.0));
        }


        /**
         * Computes the footpoint latitude for use in converting transverse Mercator coordinates to ellipsoidal coordinates.
         *
         * @param y       The UTM northing coordinate, in meters.
         * @return        The footpoint latitude, in radians.
         */
        private static double footpointLatitude(double y)
        {
            double y_, alpha_, beta_, gamma_, delta_, epsilon_, n;

            //precalculations
            n = (sm_a - sm_b) / (sm_a + sm_b);
            alpha_ = ((sm_a + sm_b) / 2.0) * (1 + (Math.Pow(n, 2.0) / 4) + (Math.Pow(n, 4.0) / 64));
            y_ = y / alpha_;
            beta_ = (3.0 * n / 2.0) + (-27.0 * Math.Pow(n, 3.0) / 32.0) + (269.0 * Math.Pow(n, 5.0) / 512.0);
            gamma_ = (21.0 * Math.Pow(n, 2.0) / 16.0) + (-55.0 * Math.Pow(n, 4.0) / 32.0);
            delta_ = (151.0 * Math.Pow(n, 3.0) / 96.0) + (-417.0 * Math.Pow(n, 5.0) / 128.0);
            epsilon_ = (1097.0 * Math.Pow(n, 4.0) / 512.0);
            return y_ + (beta_ * Math.Sin(2.0 * y_)) + (gamma_ * Math.Sin(4.0 * y_)) + (delta_ * Math.Sin(6.0 * y_)) + (epsilon_ * Math.Sin(8.0 * y_));
        }


        /**
         * Converts a latitude/longitude pair to x and y coordinates in the Transverse Mercator projection.  Note that Transverse Mercator is not
         * the same as UTM; a scale factor is required to convert between them.
         * 
         * @param phi      Latitude of the point, in radians.
         * @param lambda   Longitude of the point, in radians.
         * @param lambda0  Longitude of the central meridian to be used, in radians.
         * @return         A 2-element array containing the x and y coordinates of the computed point.
         */
        private static double[] mapLatLonToXY(double phi, double lambda, double lambda0)
        {
            double[] xy = new double[2];
            double N, nu2, ep2, t, t2, l;
            double l3coef, l4coef, l5coef, l6coef, l7coef, l8coef;

            //precalculations
            ep2 = (Math.Pow(sm_a, 2.0) - Math.Pow(sm_b, 2.0)) / Math.Pow(sm_b, 2.0);
            nu2 = ep2 * Math.Pow(Math.Cos(phi), 2.0);
            N = Math.Pow(sm_a, 2.0) / (sm_b * Math.Sqrt(1 + nu2));
            t = Math.Tan(phi);
            t2 = t * t;
            l = lambda - lambda0;

            //Precalculate coefficients for l**n in the equations below so a normal human being can read the expressions for easting and northing
            //-- l**1 and l**2 have coefficients of 1.0
            l3coef = 1.0 - t2 + nu2;
            l4coef = 5.0 - t2 + 9 * nu2 + 4.0 * (nu2 * nu2);
            l5coef = 5.0 - 18.0 * t2 + (t2 * t2) + 14.0 * nu2 - 58.0 * t2 * nu2;
            l6coef = 61.0 - 58.0 * t2 + (t2 * t2) + 270.0 * nu2 - 330.0 * t2 * nu2;
            l7coef = 61.0 - 479.0 * t2 + 179.0 * (t2 * t2) - (t2 * t2 * t2);
            l8coef = 1385.0 - 3111.0 * t2 + 543.0 * (t2 * t2) - (t2 * t2 * t2);

            // Calculate easting (x)
            xy[0] = N * Math.Cos(phi) * l
                + (N / 6.0 * Math.Pow(Math.Cos(phi), 3.0) * l3coef * Math.Pow(l, 3.0))
                + (N / 120.0 * Math.Pow(Math.Cos(phi), 5.0) * l5coef * Math.Pow(l, 5.0))
                + (N / 5040.0 * Math.Pow(Math.Cos(phi), 7.0) * l7coef * Math.Pow(l, 7.0));

            //Calculate northing (y)
            xy[1] = arcLengthOfMeridian(phi)
                + (t / 2.0 * N * Math.Pow(Math.Cos(phi), 2.0) * Math.Pow(l, 2.0))
                + (t / 24.0 * N * Math.Pow(Math.Cos(phi), 4.0) * l4coef * Math.Pow(l, 4.0))
                + (t / 720.0 * N * Math.Pow(Math.Cos(phi), 6.0) * l6coef * Math.Pow(l, 6.0))
                + (t / 40320.0 * N * Math.Pow(Math.Cos(phi), 8.0) * l8coef * Math.Pow(l, 8.0));

            return xy;
        }


        /**
         * Converts x and y coordinates in the Transverse Mercator projection to a latitude/longitude pair.  Note that Transverse Mercator is not
         * the same as UTM; a scale factor is required to convert between them.
         * 
         * @param x        The easting of the point, in meters.
         * @param y        The northing of the point, in meters.
         * @param lambda0  Longitude of the central meridian to be used, in radians.
         * @return         A 2-element containing the latitude and longitude in radians.
         */
        private static double[] mapXYToLatLon(double x, double y, double lambda0)
        {
            double phif, Nf, Nfpow, nuf2, ep2, tf, tf2, tf4, cf;
            double x1frac, x2frac, x3frac, x4frac, x5frac, x6frac, x7frac, x8frac;
            double x2poly, x3poly, x4poly, x5poly, x6poly, x7poly, x8poly;
            double[] philambda = new double[2];

            //Get the value of phif, the footpoint latitude
            phif = footpointLatitude(y);

            //Precalculations
            ep2 = (Math.Pow(sm_a, 2.0) - Math.Pow(sm_b, 2.0)) / Math.Pow(sm_b, 2.0);
            cf = Math.Cos(phif);
            nuf2 = ep2 * Math.Pow(cf, 2.0);
            Nf = Math.Pow(sm_a, 2.0) / (sm_b * Math.Sqrt(1 + nuf2));
            Nfpow = Nf;
            tf = Math.Tan(phif);
            tf2 = tf * tf;
            tf4 = tf2 * tf2;

            //Precalculate fractional coefficients for x**n in the equations below to simplify the expressions for latitude and longitude.
            x1frac = 1.0 / (Nfpow * cf);
            Nfpow *= Nf;   							//now equals Nf**2)
            x2frac = tf / (2.0 * Nfpow);
            Nfpow *= Nf;							//now equals Nf**3)
            x3frac = 1.0 / (6.0 * Nfpow * cf);
            Nfpow *= Nf;							//now equals Nf**4)
            x4frac = tf / (24.0 * Nfpow);
            Nfpow *= Nf;							//now equals Nf**5)
            x5frac = 1.0 / (120.0 * Nfpow * cf);
            Nfpow *= Nf;							//now equals Nf**6)
            x6frac = tf / (720.0 * Nfpow);
            Nfpow *= Nf;   							//now equals Nf**7)
            x7frac = 1.0 / (5040.0 * Nfpow * cf);
            Nfpow *= Nf;   							//now equals Nf**8)
            x8frac = tf / (40320.0 * Nfpow);

            //Precalculate polynomial coefficients for x**n. x**1 does not have a polynomial coefficient
            x2poly = -1.0 - nuf2;
            x3poly = -1.0 - 2 * tf2 - nuf2;
            x4poly = 5.0 + 3.0 * tf2 + 6.0 * nuf2 - 6.0 * tf2 * nuf2 - 3.0 * (nuf2 * nuf2) - 9.0 * tf2 * (nuf2 * nuf2);
            x5poly = 5.0 + 28.0 * tf2 + 24.0 * tf4 + 6.0 * nuf2 + 8.0 * tf2 * nuf2;
            x6poly = -61.0 - 90.0 * tf2 - 45.0 * tf4 - 107.0 * nuf2 + 162.0 * tf2 * nuf2;
            x7poly = -61.0 - 662.0 * tf2 - 1320.0 * tf4 - 720.0 * (tf4 * tf2);
            x8poly = 1385.0 + 3633.0 * tf2 + 4095.0 * tf4 + 1575 * (tf4 * tf2);

            //Calculate latitude
            philambda[0] = phif + x2frac * x2poly * (x * x)
                + x4frac * x4poly * Math.Pow(x, 4.0)
                + x6frac * x6poly * Math.Pow(x, 6.0)
                + x8frac * x8poly * Math.Pow(x, 8.0);

            //Calculate longitude
            philambda[1] = lambda0 + x1frac * x
                + x3frac * x3poly * Math.Pow(x, 3.0)
                + x5frac * x5poly * Math.Pow(x, 5.0)
                + x7frac * x7poly * Math.Pow(x, 7.0);

            return philambda;
        }



        /**
         * Converts a latitude/longitude pair to x and y coordinates in the Universal Transverse Mercator projection.
         * 
         * @param lat      Latitude of the point, in degrees.
         * @param lon      Longitude of the point, in degrees.
         * 
         * @return         A 2-element array where the UTM x and y values will be stored.
         */
        public static double[] latLonToUTMXY(double lat, double lon)
        {
            int zone = (int)Math.Floor((lon + 180.0) / 6) + 1;
            double[] xy = mapLatLonToXY(ToRadians(lat), ToRadians(lon), UTMCentralMeridian(zone));

            //Adjust easting and northing for UTM system.
            xy[0] = xy[0] * UTMScaleFactor + 500000.0;
            xy[1] = xy[1] * UTMScaleFactor;
            if (xy[1] < 0.0)
                xy[1] = xy[1] + 10000000.0;

            return xy;
        }


        /**
         * Converts x and y coordinates in the Universal Transverse Mercator projection to a latitude/longitude pair.
         * 
         * @param x          The easting of the point, in meters.
         * @param y          The northing of the point, in meters.
         * @param zone       The UTM zone in which the point lies.
         * @param southhemi  True if the point is in the southern hemisphere, false otherwise.
         * @return           A 2-element array containing the latitude and longitude of the point, in radians.
         */
        public static double[] UTMXYToLatLon(double x, double y, int zone, bool southhemi)
        {
            x -= 500000.0;
            x /= UTMScaleFactor;

            if (southhemi)				//If in southern hemisphere, adjust y accordingly.
                y -= 10000000.0;
            y /= UTMScaleFactor;

            return mapXYToLatLon(x, y, UTMCentralMeridian(zone));
        }
        private static double ToRadians(double degrees)
        {
            return ((degrees * Math.PI) / 180);
        }

    }


}
